/*
 * script from BSD
 */

/* 
 * If you have some troubles with SGTTY = 0, change SGTTY = 1 
 */
#define SGTTY 0


/* default values */
int	ks2n = 1;	/* ksc5601 on nbyte code */
int	h1e1 = 1;	/* hangul size is equal to alphabet size */

#define KSEMUL "ksemul"	/* default name for ke emulstor on n byte */
#define NEMUL "nemul"


/* 
 * this changes fluchchar with '^P' because the default value, ^O is used as
 * hangul end. (default of AT&T : ('o'&037))
 */
#define FLUSHCHAR ('p'&037)

#include <sys/types.h>
#include <sys/stat.h>

#if SGTTY
#include <sgtty.h>
#else
#include <termios.h>
#endif
/*
#include <sys/ioctl.h>
*/

#ifndef O_RAW
#define O_RAW RAW
#define O_CBREAK CBREAK
#define O_ECHO ECHO
#endif


#include <sys/time.h>
#include <sys/file.h>
#include <sys/signal.h>
#include <stdio.h>
/*
#include <paths.h>
*/
#define _PATH_BSHELL	"/bin/sh"
#define _PATH_TTY	"/dev/tty"

char	*shell;
FILE	*fscript;
int	master;
int	slave;
int	child;
int	subchild;
char	*fname;
char 	ptyname[32];
char 	ttyname[32];
int	tslot = 0;


#if SGTTY
struct	sgttyb _tty;
#else
struct	termios tt;
#endif

struct	winsize win;
int	lb;
int	l;
char	line[] = "/dev/ptyXX";
int	aflg;
int	scr = 0;


main(argc, argv)
	int argc;
	char *argv[];
{
	extern char *optarg;
	extern int optind;
	int ch;
	void finish();
	char *getenv();

	if (strcmp(argv[0], KSEMUL) == 0) {
	    ks2n = 1;
	}
	else if (strcmp(argv[0], NEMUL) == 0) {
	    ks2n = 0;
	}

	while ((ch = getopt(argc, argv, "asknh?12")) != EOF)
		switch((char)ch) {
		case 'a':
			aflg++;
			break;
		case 's':
			scr = 1;
			break;
		case 'k':
			ks2n = 1;
			break;
		case 'n':
			ks2n = 0;
			break;
		case '1':
			h1e1 = 1;
			break;
		case '2':
			h1e1 = 0;
			break;
		case '?':
		case 'h':
		default:
			fprintf(stderr, "HEMUL 0.9.2 hangul code emulator. 1992.4.2\n");
			fprintf(stderr, 
			    "usage: %s [-k|-n] [-1|-2] [-a] [-s file]\n",
			    argv[0]);
			fprintf(stderr, "-k : ks emulator on n-byte hangul\n");
			fprintf(stderr, "-n : n-byte emulator on ks hangul\n");
			fprintf(stderr, "-1,2 : the size of a hangul char.\n");
			fprintf(stderr, "-s : log the script in the file\n");
			fprintf(stderr, "-a : log the script appending to the file\n");
			fprintf(stderr, "default : -k -1\n");
			exit(1);
		}
	argc -= optind;
	argv += optind;

	if (argc > 0)
		fname = argv[0];
	else
		fname = "typescript";
	if (scr && (fscript = fopen(fname, aflg ? "a" : "w")) == NULL) {
		perror(fname);
		fail();
	}

	shell = getenv("SHELL");
	if (shell == NULL)
		shell = _PATH_BSHELL;

	getmaster();
	if (ks2n)
	printf("KSC5601 hangul emulator for N-byte hangul terminal started!\n");
	else
	printf("N-byte hangul emulator for KSC5601 hangul terminal started!\n");
	if (scr) printf("Script started, file is %s\n", fname);
	fixtty();

	(void) signal(SIGCHLD, finish);
	child = fork();
	if (child < 0) {
		perror("fork");
		fail();
	}
	if (child == 0) {
			doshell();
	}
	else {
		InitUtmp();
		doinputoutput();
	}
}


#include <sys/wait.h>

void
finish()
{
	union wait status;
	register int pid;
	register int die = 0;

	while ((pid = wait3((int *)&status, WNOHANG, 0)) > 0)
		if (pid == child)
			die = 1;

	if (die)
		done();
}

doinputoutput()
{
	register int cc;
	char buf[BUFSIZ];

	fd_set fdset;
	time_t tvec, time();
	char *ctime();

	tvec = time((time_t *)NULL);
	if (scr) fprintf(fscript, "Script started on %s", ctime(&tvec));

	while (1) {
	    FD_ZERO(&fdset);
	    FD_SET(0, &fdset);
	    FD_SET(master, &fdset);
	    if (select( master+1, &fdset, NULL, NULL, NULL ) < 0) break;

	    if (FD_ISSET(0, &fdset)) {
		cc = hread(0, buf, BUFSIZ);
		if (cc < 0) break;

		(void) write(master, buf, cc);
	    }
	    if (FD_ISSET(master, &fdset)) {
		cc = read(master, buf, sizeof (buf));
		if (cc <= 0) break;

		(void) hwrite(1, buf, cc);
		if (scr) (void) fwrite(buf, 1, cc, fscript);
	    }
	}
	done();
}

doshell()
{
	int t;

	/***/
	t = open(_PATH_TTY, O_RDWR);
	if (t >= 0) {
		(void) ioctl(t, TIOCNOTTY, (char *)0);
		(void) close(t);
	}
	/***/
	getslave();
	(void) close(master);
	if (scr) (void) fclose(fscript);
	(void) dup2(slave, 0);
	(void) dup2(slave, 1);
	(void) dup2(slave, 2);
	(void) close(slave);

	execl(shell, "sh", "-i", 0);
	perror(shell);
	fail();
}

fixtty()
{
#if SGTTY
	struct sgttyb tty;
	int flags;

	tty = _tty;
	tty.sg_flags |= O_RAW|O_CBREAK;
	tty.sg_flags &= ~O_ECHO;
	stty(0, &tty);

	flags = LPASS8;
	ioctl(slave, TIOCLBIS, &flags);
	flags = LCTLECH;
	ioctl(slave, TIOCLBIC, &flags);

#else
	struct termios rtt;
	int i;

	rtt = tt;
/*	cfmakeraw(&rtt); */
	for ( i = 0; i < NCCS; i++ ) rtt.c_cc[i] = NULL;
	rtt.c_cc[VMIN] = 255;
	rtt.c_cc[VTIME] = 5;
	rtt.c_iflag &= ~ISTRIP;
	rtt.c_lflag &= ~ICANON & ~ECHO & ~ECHOCTL;
	(void) ioctl(0, TCSETSF, &rtt);
#endif
}

fail()
{

	(void) kill(0, SIGTERM);
	done();
}

done()
{
	time_t tvec, time();
	char *ctime();

	tvec = time((time_t *)NULL);
	if (scr) {
	    fprintf(fscript,"\nScript done on %s", ctime(&tvec));
	    (void) fclose(fscript);
	}

	RemoveUtmp(tslot);

	(void) close(master);
#if SGTTY
		stty(0, &_tty);
#else
/*
		(void) tcsetattr(0, TCSAFLUSH, &tt);
*/
	     (void) ioctl(0, TCSETS, &tt);
#endif
	if (scr) printf("Script done, file is %s", fname);
	printf("\nEmulation is ended. Thank you for using me. ");
	printf("(by hanmaum) 1992.4.2\n");
	exit(0);
}

getmaster()
{
	char *pty, *bank, *cp;
	struct stat stb;

	pty = &line[strlen("/dev/ptyp")];
	for (bank = "pqrs"; *bank; bank++) {
		line[strlen("/dev/pty")] = *bank;
		*pty = '0';
		if (stat(line, &stb) < 0)
			break;
		for (cp = "0123456789abcdef"; *cp; cp++) {
			*pty = *cp;
			master = open(line, O_RDWR);
			if (master >= 0) {
				char *tp = &line[strlen("/dev/")];
				int ok;

				/* verify slave side is usable */
				*tp = 't';
				strcpy(ttyname, line);
				ok = access(line, R_OK|W_OK) == 0;
				*tp = 'p';
				strcpy(ptyname, line);
				if (ok) {
#if SGTTY
					(void) gtty(0, &_tty);
#else
/*
					(void) tcgetattr(0, &tt);
*/
					(void) ioctl(0, TCGETS, &tt);
#endif

				    	(void) ioctl(0, TIOCGWINSZ, 
						(char *)&win);

					chown(ttyname, getuid(), getgid());
					tslot = SetUtmp(ttyname);

					return;
				}
				(void) close(master);
			}
		}
	}
	fprintf(stderr, "Out of pty's\n");
	fail();
}

getslave()
{

	line[strlen("/dev/")] = 't';
	slave = open(line, O_RDWR);
	if (slave < 0) {
		perror(line);
		fail();
	}
#if SGTTY
	(void) stty(slave, &_tty);
	{
	    struct ltchars ltc;
		int flags;

		flags = LPASS8;
	ioctl(slave, TIOCLBIS, &flags);
		flags = LCTLECH;
	ioctl(slave, TIOCLBIC, &flags);

	    ioctl(slave, TIOCGLTC, &ltc);
	    ltc.t_flushc = FLUSHCHAR;
	    ioctl(slave, TIOCSLTC, &ltc);
	}
#else
	{
	    struct termios rtt;
	    rtt = tt;
	    rtt.c_cc[VDISCARD] = FLUSHCHAR;
	    rtt.c_iflag &= ~ISTRIP;
	    rtt.c_lflag &= ~ECHOCTL;
/*
	    (void) tcsetattr(slave, TCSAFLUSH, &rtt);
*/
	     (void) ioctl(slave, TCSETSF, &rtt);
	}
	(void) ioctl(slave, TIOCSWINSZ, (char *)&win);
/*
	(void) setsid();
	(void) ioctl(slave, TIOCSCTTY, 0);
*/
#endif
}

#define LASTLOG 0
#define WTMP 0

char UtmpName[] = "/dev/utmp";
char etc_wtmp[] = "/usr/adm/wtmp";
char etc_lastlog[] = "/usr/adm/lastlog";

#include <pwd.h>
#include <utmp.h>
#include <errno.h>
#include <malloc.h>

extern char *rindex();
extern char *getlogin();

int utmpf;
int utmp = 0;
char *LoginName;
static char PtyName[32], TtyName[32];

#ifdef GETTTYENT
#   include <ttyent.h>
#else
    static struct ttyent {
        char *ty_name;
    } *getttyent();
    static char *ttcur, *ttnext;
    static char ttys[] = "/etc/ttys";
#endif



Msg(err,nam)
int err;
char nam[];
{
    printf("%d %s\n", err, nam);
}

static InitUtmp () {
    struct passwd *p;

    if ((utmpf = open (UtmpName, O_WRONLY)) == -1) {
        if (errno != EACCES)
            Msg (errno, UtmpName);
        return;
    }
    if ((LoginName = getlogin ()) == 0 || LoginName[0] == '\0') {
        if ((p = getpwuid (getuid ())) == 0)
            return;
        LoginName = p->pw_name;
    }
    utmp = 1;
}

static SetUtmp (name) char *name; {
    register char *p;
    register struct ttyent *tp;
    register slot = 1;
    struct utmp u;

    if (!utmp)
        return 0;
    if (p = rindex (name, '/'))
        ++p;
    else p = name;
    setttyent ();
    while ((tp = getttyent ()) != NULL && strcmp (p, tp->ty_name) != 0)
        ++slot;
    if (tp == NULL)
        return 0;
    strncpy (u.ut_line, p, 8);
    strncpy (u.ut_name, LoginName, 8);
    u.ut_host[0] = '\0';
    time (&u.ut_time);
    (void) lseek (utmpf, (long)(slot * sizeof (u)), 0);
    (void) write (utmpf, (char *)&u, sizeof (u));
    return slot;
}

static RemoveUtmp (slot) {
    struct utmp u;

    if (slot) {
        bzero ((char *)&u, sizeof (u));
        (void) lseek (utmpf, (long)(slot * sizeof (u)), 0);
        (void) write (utmpf, (char *)&u, sizeof (u));
    }
}


#ifndef GETTTYENT

static setttyent () {
    struct stat s;
    register f;
    register char *p, *ep;

    if (ttnext) {
        ttnext = ttcur;
        return;
    }
    if ((f = open (ttys, O_RDONLY)) == -1 || fstat (f, &s) == -1)
        Msg (errno, ttys);
    if ((ttcur = malloc (s.st_size + 1)) == 0)
        Msg (0, "Out of memory.");
    if (read (f, ttcur, s.st_size) != s.st_size)
        Msg (errno, ttys);
    close (f);
    for (p = ttcur, ep = p + s.st_size; p < ep; ++p)
        if (*p == '\n') *p = '\0';
    *p = '\0';

    ttnext = ttcur;
}

static struct ttyent *getttyent () {
    static struct ttyent t;

    if (*ttnext == '\0')
        return NULL;
    t.ty_name = ttnext + 2;
    ttnext += strlen (ttnext) + 1;
    return &t;
}

#endif







int hread (fd, buf, siz)
int fd;
char *buf;
int siz;
{
    return (ks2n) ? hnread(fd, buf, siz) : hksread(fd, buf, siz);
}

int hwrite (fd, buf, siz)
int fd;
char *buf;
int siz;
{
    return (ks2n) ? hnwrite(fd, buf, siz) : hkswrite(fd, buf, siz);
}
